@@ -15,6 +15,9 @@
 #include <ctype.h>
 #include <signal.h>
 #include <time.h>
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <fcntl.h>
 #include "cgi.h"
 
 char *cgi_query=(char*)0;
@@ -539,6 +542,11 @@ cgi_pclose_fix()
   sigemptyset(&mysigmask);
   sigaddset(&mysigmask, SIGCHLD);
   sigprocmask(SIG_UNBLOCK, &mysigmask, 0);
+
+  /* Under Irix 5.3 we can be killed if sendmail died prematurely -brlewis */
+  sigemptyset(&mysigmask);
+  sigaddset(&mysigmask, SIGPIPE);
+  sigprocmask(SIG_BLOCK, &mysigmask, 0);
 #else
   /* Hopefully this works for old versions of SunOS.  We'll see. */
   signal(SIGCHLD, SIG_DFL);
@@ -550,11 +558,12 @@ cgi_mail_template(formp, templatefile)
      cgi_form *formp;
      char *templatefile;
 {
-  FILE *mypipe;
+  FILE *mypipe, *errorfp;
   int retval;
-  char buf[BUFSIZ];
+  char buf[BUFSIZ], command[BUFSIZ];
   int nbytes;
-  char *fromhost, *servername, datebuf[80];
+  char *fromhost, *servername, datebuf[80], errorfile[L_tmpnam];
+  int old_stdout, old_stderr, errorfd;
   time_t rightnow;
   struct tm *timeptr;
 
@@ -562,10 +571,36 @@ cgi_mail_template(formp, templatefile)
   retval=cgi_template_fill(formp, templatefile);
   if (retval) return(retval);
 
+  /* Get a temporary file for error messages */
+  tmpnam(errorfile);
+  errorfd=open(errorfile, O_WRONLY|O_CREAT|O_EXCL, 0644);
+  if (errorfd != -1)
+    {
+      old_stdout=dup(1);
+      dup2(errorfd, 1);
+      old_stderr=dup(2);
+      dup2(errorfd, 2);
+    }      
+
   /* open pipe to sendmail */
   cgi_pclose_fix();
-  strcat(strcpy(buf, PATH_SENDMAIL), " -oi -t");
-  if (!(mypipe=popen(buf, "w")))
+  strcat(strcpy(command, PATH_SENDMAIL), " -oi -t");
+  if (strstr(cgi_value(formp, "cgiemail-mailopt"), "sync"))
+    strcat(command, " -odi");	/* synchronous delivery */
+  mypipe = popen(command, "w");
+
+  /* Restore stdout/stderr */
+  if (errorfd != -1)
+    {
+      dup2(old_stdout, 1);
+      close(old_stdout);
+      dup2(old_stderr, 2);
+      close(old_stderr);
+      close(errorfd);
+    }
+
+  /* Check for failed popen */
+  if (!mypipe)
     {
       formp->errcond=1;
       strcpy(formp->errmsg, "500 Could not open sendmail pipe.");
@@ -607,14 +642,35 @@ cgi_mail_template(formp, templatefile)
 
   /* close */
   retval=pclose(mypipe);
-  if (retval)
+
+  /*
+   * The return value of pclose is not always meaningful.
+   * However, sendmail does not write to stdout/stderr if successful.
+   */
+
+  /* If we can't open the errorfile, something went wrong. */
+  errorfp = fopen(errorfile, "r");
+  if (!errorfp)
     {
       formp->errcond=1;
-      sprintf(formp->errmsg, "500 Exit status of sendmail: %d", retval);
-      cgi_concat_errno(formp->errmsg);
-      return(1);
+      sprintf(formp->errmsg,
+	     "500 sendmail exit %d - check httpd error logs", retval);
+      strcpy(formp->errinfo, command);
     }
-  return(0);
+  else
+    {
+      /* See if there was any error message */
+      fread(formp->errinfo, sizeof(char), CGI_ERRMSG_MAX, errorfp);
+      if (cgi_nonblank(formp->errinfo))
+	{
+	  formp->errcond=1;
+	  sprintf(formp->errmsg, "500 sendmail exit %d with error message",
+		 retval);
+	}
+    }
+
+  unlink(errorfile);
+  return(formp->errcond);
 }
 
 void
@@ -626,17 +682,44 @@ cgi_output_failure(formp, msg)
   printf("Status: %s\r\n\r\n", formp->errmsg);
   puts("<HEAD><TITLE>Error</TITLE></HEAD>");
   printf("<BODY><H1>Error</H1>%s<P>", msg);
-  puts("<SAMP>");
+  puts("<BLOCKQUOTE><STRONG><SAMP>");
   puts(formp->errmsg);
-  puts("</SAMP><P>");
+  puts("</SAMP></STRONG><P>");
+  if (*(formp->errinfo))
+    {
+      puts("<PRE>");
+      puts(formp->errinfo);
+      puts("</PRE>");
+    }
+  puts("</BLOCKQUOTE>");
   (void) cgi_output_value(formp, CGI_ADDENDUM, NULL, stdout);
   puts("<P><EM>cgiemail ");
   puts(CGIEMAIL_RELEASE);
   puts("</EM></BODY>");
-  puts("</BODY>");
   return;
 }
 
+/* Function to translate an integer character into an HTML entity string */
+char *
+cgi_char2entity(c)
+     int c;
+{
+  static char retval[2];
+
+  switch (c)
+    {
+    case (int)'&':
+      return("&amp;");
+    case (int)'<':
+      return("&lt;");
+    case (int)'>':
+      return("&gt;");
+    default:
+      sprintf(retval, "%c", c);
+      return(retval);
+    }
+}
+  
 void
 cgi_ascii2html(instream, outstream)
      FILE *instream, *outstream;
@@ -644,21 +727,7 @@ cgi_ascii2html(instream, outstream)
   int c;
 
   while ((c=fgetc(instream)) != EOF)
-    switch (c)
-      {
-      case (int)'&':
-	fputs("&amp;", outstream);
-	break;
-      case (int)'<':
-	fputs("&lt;", outstream);
-	break;
-      case (int)'>':
-	fputs("&gt;", outstream);
-	break;
-      default:
-	fputc(c, outstream);
-	break;
-      }
+    fputs(cgi_char2entity(c), outstream);
   return;
 }
 
@@ -711,10 +780,19 @@ int
 cgi_standard_email()
 {
   cgi_form form;
+  char *template_filename;
+
+  template_filename = getenv("PATH_TRANSLATED");
+  if (!template_filename || !(*template_filename))
+    {
+      /* FIXME: shouldn't hard-code this URL */
+      cgi_redirect("http://web.mit.edu/wwwdev/cgiemail/nopath.html");
+      return(1);
+    }
 
   if (cgi_alloc_form(&form) ||
       cgi_parse_form(&form) ||
-      cgi_mail_template(&form, getenv("PATH_TRANSLATED")))
+      cgi_mail_template(&form, template_filename))
     {
       cgi_output_failure(&form, "No email was sent due to an error.");
       cgi_free_form(&form);
@@ -730,10 +808,19 @@ int
 cgi_standard_echo()
 {
   cgi_form form;
+  char *template_filename;
+
+  template_filename = getenv("PATH_TRANSLATED");
+  if (!template_filename || !(*template_filename))
+    {
+      /* FIXME: shouldn't hard-code this URL */
+      cgi_redirect("http://web.mit.edu/wwwdev/cgiemail/nopath.html");
+      return(1);
+    }
 
   if (cgi_alloc_form(&form) ||
       cgi_parse_form(&form) ||
-      cgi_template_fill(&form, getenv("PATH_TRANSLATED")))
+      cgi_template_fill(&form, template_filename))
     {
       cgi_output_failure(&form, "Form was not processed due to an error.");
       cgi_free_form(&form);
